package FaDe;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * The FeatureCalculator class is used to calculate features describing user's behaviour in using mouse and keyboard.
 * First the class has to be filled with data by calling either the <code>collectData</code> or the <code>useDataFromLogger</code> methods.
 * Than any feature can be obtained with the <code>getFeature(feature>)</code> method, or the whole set of features can be obtained with the <code>getFeatureSet()</code> method.
 * 
 * @author Giorgio Luparia
 */
public class FeatureCalculator {
    /*
	 * Event logger
	 */
	private Logger logger;
    
    /*
     * Feature set
     */
	private FeatureSet featureSet;
	
	/*
	 * Calculated values cache
	 */
	private BigInteger totalKeyStrokeDuration;

	
	/**
	 * Class constructor.
	 */
	public FeatureCalculator() {
		logger = new Logger();
    }
    
	/**
	 * Causes the class to collect user input data for the default time.
	 * The method is blocking.
	 */
	public void collectData() {
		collectData(Config.dataCollectionTime);
	}
	
	/**
	 * Causes the class to collect user input data for the specified minutes.
	 * The method is blocking.
	 *
	 * @param	minutes	the minutes to collect data for
	 */
	public void collectData(int minutes) {
		featureSet = new FeatureSet();
		totalKeyStrokeDuration = null;
		logger.run(minutes);
	}
	
	/**
	 * Adds a previously collected log to be used for feature calculation.
	 * This method is alternative to <code>collectData</code>.
	 * 
	 * @param	logger	the previously collected log
	 */
	public void useDataFromLogger(Logger logger) {
		featureSet = new FeatureSet();
		totalKeyStrokeDuration = null;
		this.logger = logger;
	}
	
	/**
	 * Calculates and returns the specified feature value.
	 * 
	 * @param	feature	the feature to return
	 * @return			the specified feature value
	 */
	public Double getFeature(FeatureSet.Feature feature) {
		calculateFeature(feature);
		return featureSet.getFeature(feature);
	}
	
	/**
	 * Calculate and returns the entire FeatureSet.
	 * 
	 * @return	the entire FeatureSet
	 */
	public FeatureSet getFeatureSet()
	{
		for (FeatureSet.Feature feature: FeatureSet.Feature.values()) {
			calculateFeature(feature);
		}
		return featureSet;
	}
	
	/**
	 * Calculates and returns the total time (in microseconds) a specified key has been pressed for.
	 * 
	 * @param	key	the key to consider
	 * @return		microseconds the key has been pressed for
	 */
	private BigInteger totalStrokeDurationOfKey(Key key) {
		BigInteger duration = BigInteger.valueOf(0L);
		for (KeyStroke i: logger.getKeyStrokeLog())
			if (i.getKey() == key)
				duration = duration.add(BigInteger.valueOf(i.getStrokeDurationInMicroseconds()));
		return duration;
	}
	
	/**
	 * Calculates and returns the total time (in microseconds) any key has been pressed for.
	 * 
	 * @return	microseconds any key has been pressed for 
	 */
	private BigInteger totalKeyStrokeDuration() {
		if (totalKeyStrokeDuration == null) {
			totalKeyStrokeDuration = BigInteger.valueOf(0L);
			for (KeyStroke i: logger.getKeyStrokeLog())
				totalKeyStrokeDuration = totalKeyStrokeDuration.add(BigInteger.valueOf(i.getStrokeDurationInMicroseconds()));	
			}
		return totalKeyStrokeDuration;
	}
	
	/**
	 * Calculates and returns the number of times the specified key has been pressed.
	 * 
	 * @param	key	the key to consider
	 * @return		the number of times the key has been pressed for
	 */
	private int countStrokesOfKey(Key key) {
		int count = 0;
		for (KeyStroke i: logger.getKeyStrokeLog())
			if (i.getKey() == key)
				count++;
		return count;
	}
	
	/**
	 * Calculates the specified feature and adds it to the feature set.
	 * 
	 * @param	feature						the feature to calculate
	 * @throws	NoDataAvailableException	if no data is available for calculation (<code>collectData</code> or <code>useDataFromLogger()</code> must be called
	 */
	private void calculateFeature(FeatureSet.Feature feature) {
		if (featureSet == null)
			throw new NoDataAvailableException();
		
		if (featureSet.getFeature(feature) == null)
		{
			List<KeyStroke> keyStrokeLog = logger.getKeyStrokeLog();
			List<MouseButtonStroke> mouseButtonStrokeLog = logger.getMouseButtonStrokeLog();
			List<MouseMovement> mousePositionLog = logger.getMousePositionLog();
			List<Event> mouseFullLog = new ArrayList<Event>();
			mouseFullLog.addAll(mouseButtonStrokeLog);
			mouseFullLog.addAll(mousePositionLog);
			Collections.sort(mouseFullLog);
			
			switch (feature) {
				case KeystrokeDurationMin:
				case KeystrokeDurationAve:
				case KeystrokeDurationMax:
				{
					Long min = Long.MAX_VALUE, max = new Long(0);
					BigInteger ave = new BigInteger("0");
					int strokeCounter = 0;
					
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
					{
						for (Iterator<KeyStroke> i = keyStrokeLog.iterator(); i.hasNext();) {
							KeyStroke tmp = i.next();
							long strokeDuration = tmp.getStrokeDurationInMicroseconds();
							if (strokeDuration < min)
								min = strokeDuration;
							if (strokeDuration > max)
								max = strokeDuration;
							ave = ave.add(new BigInteger(Long.toString(strokeDuration)));
							strokeCounter++;
						}
						if (strokeCounter > 0)
						{
							ave = ave.divide(new BigInteger(Integer.toString(strokeCounter)));
							featureSet.setFeature(FeatureSet.Feature.KeystrokeDurationMin, min);
							featureSet.setFeature(FeatureSet.Feature.KeystrokeDurationAve, ave.doubleValue());
							featureSet.setFeature(FeatureSet.Feature.KeystrokeDurationMax, max);
						}
					}
					return;
				}
				case WritingSpeedMin:
				case WritingSpeedAve:
				case WritingSpeedMax:
				{
					Double freq = 0D;
					Double min = Double.MAX_VALUE, max = new Double(0);
					BigDecimal ave = new BigDecimal("0");
					
					Instant sessionStart = null;
					Instant lastTimestamp = null;
					int strokeCounter = 0;
					int sessionCounter = 0;
					
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
					{
						for (KeyStroke stroke: keyStrokeLog) {
							if (sessionStart == null || stroke.getInstant().minus(lastTimestamp) > Config.writingSessionLimit) {
								if (strokeCounter >= Config.writingSessionMinLength) {
									freq = (double) strokeCounter / (lastTimestamp.minus(sessionStart));
									if (freq < min)
										min = freq;
									if (freq > max)
										max = freq;
									ave = ave.add(new BigDecimal(freq));
									sessionCounter++;
								}
								sessionStart = stroke.getInstant();
								lastTimestamp = stroke.getInstant();
								strokeCounter = 1;
							} else {
								strokeCounter++;
								lastTimestamp = stroke.getInstant();
							}
						}
						if (strokeCounter >= Config.writingSessionMinLength) {
							freq = (double) strokeCounter / (lastTimestamp.minus(sessionStart));
							if (freq < min)
								min = freq;
							if (freq > max)
								max = freq;
							ave = ave.add(new BigDecimal(freq));
							sessionCounter++;
						}
						if (sessionCounter > 0) {
							ave = ave.divide(new BigDecimal(sessionCounter), RoundingMode.HALF_EVEN);
							featureSet.setFeature(FeatureSet.Feature.WritingSpeedMin, min);
							featureSet.setFeature(FeatureSet.Feature.WritingSpeedAve, ave.doubleValue());
							featureSet.setFeature(FeatureSet.Feature.WritingSpeedMax, max);
						}
					}
					return;
				}
				case RelativeFrequencyOfKeyENTER:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeFrequencyOfKeyENTER, (((double) countStrokesOfKey(Key.ENTER)) / keyStrokeLog.size()));
					return;
				case RelativeFrequencyOfKeySPACEBAR:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeFrequencyOfKeySPACEBAR, (((double) countStrokesOfKey(Key.SPACEBAR)) / keyStrokeLog.size()));
					return;
				case RelativeFrequencyOfKeyTAB:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeFrequencyOfKeyTAB, (((double) countStrokesOfKey(Key.TAB)) / keyStrokeLog.size()));
					return;
				case RelativeFrequencyOfKeyCANCEL:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeFrequencyOfKeyCANCEL, (((double) countStrokesOfKey(Key.CANCEL)) / keyStrokeLog.size()));
					return;
				case RelativeFrequencyOfKeyBACKSPACE:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeFrequencyOfKeyBACKSPACE, (((double) countStrokesOfKey(Key.BACKSPACE)) / keyStrokeLog.size()));
					return;
				case RelativeFrequencyOfKeyARROW:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeFrequencyOfKeyARROW, (((double) countStrokesOfKey(Key.ARROW)) / keyStrokeLog.size()));
					return;
				case RelativeFrequencyOfKeyCONTROL:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeFrequencyOfKeyCONTROL, (((double) countStrokesOfKey(Key.CONTROL)) / keyStrokeLog.size()));
					return;
				case RelativeFrequencyOfKeySHIFT:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeFrequencyOfKeySHIFT, (((double) countStrokesOfKey(Key.SHIFT)) / keyStrokeLog.size()));
					return;
				case RelativeFrequencyOfKeyALT:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeFrequencyOfKeyALT, (((double) countStrokesOfKey(Key.ALT)) / keyStrokeLog.size()));
					return;
				case RelativeDurationOfKeyENTER:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeDurationOfKeyENTER, (new BigDecimal(totalStrokeDurationOfKey(Key.ENTER))).divide(new BigDecimal(totalKeyStrokeDuration()), 20, RoundingMode.HALF_DOWN).doubleValue());
					return;
				case RelativeDurationOfKeySPACEBAR:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeDurationOfKeySPACEBAR, (new BigDecimal(totalStrokeDurationOfKey(Key.SPACEBAR))).divide(new BigDecimal(totalKeyStrokeDuration()), 20, RoundingMode.HALF_DOWN).doubleValue());
					return;
				case RelativeDurationOfKeyTAB:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeDurationOfKeyTAB, (new BigDecimal(totalStrokeDurationOfKey(Key.TAB))).divide(new BigDecimal(totalKeyStrokeDuration()), 20, RoundingMode.HALF_DOWN).doubleValue());
					return;
				case RelativeDurationOfKeyCANCEL:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeDurationOfKeyCANCEL, (new BigDecimal(totalStrokeDurationOfKey(Key.CANCEL))).divide(new BigDecimal(totalKeyStrokeDuration()), 20, RoundingMode.HALF_DOWN).doubleValue());
					return;
				case RelativeDurationOfKeyBACKSPACE:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeDurationOfKeyBACKSPACE, (new BigDecimal(totalStrokeDurationOfKey(Key.BACKSPACE))).divide(new BigDecimal(totalKeyStrokeDuration()), 20, RoundingMode.HALF_DOWN).doubleValue());
					return;
				case RelativeDurationOfKeyARROW:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeDurationOfKeyARROW, (new BigDecimal(totalStrokeDurationOfKey(Key.ARROW))).divide(new BigDecimal(totalKeyStrokeDuration()), 20, RoundingMode.HALF_DOWN).doubleValue());
					return;
				case RelativeDurationOfKeyCONTROL:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeDurationOfKeyCONTROL, (new BigDecimal(totalStrokeDurationOfKey(Key.CONTROL))).divide(new BigDecimal(totalKeyStrokeDuration()), 20, RoundingMode.HALF_DOWN).doubleValue());
					return;
				case RelativeDurationOfKeySHIFT:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeDurationOfKeySHIFT, (new BigDecimal(totalStrokeDurationOfKey(Key.SHIFT))).divide(new BigDecimal(totalKeyStrokeDuration()), 20, RoundingMode.HALF_DOWN).doubleValue());
					return;
				case RelativeDurationOfKeyALT:
					if (keyStrokeLog != null && keyStrokeLog.size() != 0)
						featureSet.setFeature(FeatureSet.Feature.RelativeDurationOfKeyALT, (new BigDecimal(totalStrokeDurationOfKey(Key.ALT))).divide(new BigDecimal(totalKeyStrokeDuration()), 20, RoundingMode.HALF_DOWN).doubleValue());
					return;
				case MouseclickDurationMin:
				case MouseclickDurationAve:
				case MouseclickDurationMax:
				{
					Long min = Long.MAX_VALUE, max = new Long(0);
					BigInteger ave = BigInteger.valueOf(0L);
					int strokeCounter = 0;
					
					if (mouseButtonStrokeLog != null && mouseButtonStrokeLog.size() != 0)
					{
						for (Iterator<MouseButtonStroke> i = mouseButtonStrokeLog.iterator(); i.hasNext();) {
							MouseButtonStroke tmp = i.next();
							
							long strokeDuration = tmp.getStrokeDurationInMicroseconds();
							if (strokeDuration < min)
								min = strokeDuration;
							if (strokeDuration > max)
								max = strokeDuration;
							ave = ave.add(new BigInteger(Long.toString(strokeDuration)));
							strokeCounter++;
						}
						if (strokeCounter > 0)
						{
							ave = ave.divide(new BigInteger(Integer.toString(strokeCounter)));
							featureSet.setFeature(FeatureSet.Feature.MouseclickDurationMin, min);
							featureSet.setFeature(FeatureSet.Feature.MouseclickDurationAve, ave.doubleValue());
							featureSet.setFeature(FeatureSet.Feature.MouseclickDurationMax, max);
						}
					}
					return;
				}
				case MouseDoubleClickDurationMin:
				case MouseDoubleClickDurationAve:
				case MouseDoubleClickDurationMax:
				{
					Long min = Long.MAX_VALUE, max = new Long(0);
					BigInteger ave = BigInteger.valueOf(0L);
					int doubleClicksCounter = 0;
					
					if (mouseButtonStrokeLog != null && mouseButtonStrokeLog.size() != 0)
					{
						Instant lastClickPressTime = null;
						
						for (Iterator<MouseButtonStroke> i = mouseButtonStrokeLog.iterator(); i.hasNext();) {
							MouseButtonStroke tmp = i.next();
							if (lastClickPressTime != null) {
								Long interClickTime = tmp.getReleaseTime().minus(lastClickPressTime);
								if (interClickTime < Config.doubleClickLimit) {
									if (interClickTime < min)
										min = interClickTime;
									if (interClickTime > max) {
										max = interClickTime;
									}
									ave = ave.add(BigInteger.valueOf(interClickTime));
									doubleClicksCounter++;
								}
							}
							lastClickPressTime = tmp.getPressTime();
						}
						if (doubleClicksCounter > 0)
						{
							ave = ave.divide(BigInteger.valueOf(doubleClicksCounter));
							featureSet.setFeature(FeatureSet.Feature.MouseDoubleClickDurationMin, min * 1000000);
							featureSet.setFeature(FeatureSet.Feature.MouseDoubleClickDurationAve, ave.doubleValue() * 1000000);
							featureSet.setFeature(FeatureSet.Feature.MouseDoubleClickDurationMax, max * 1000000);
						}
					}
					return;
				}
				case MouseSpeedMin:
				case MouseSpeedAve:
				case MouseSpeedMax:
				case MouseMovementDeviationMin:
				case MouseMovementDeviationAve:
				case MouseMovementDeviationMax:
				case MouseHesitationsPerMovementMin:
				case MouseHesitationsPerMovementAve:
				case MouseHesitationsPerMovementMax:
				{
					Double mouseSpeedMin = Double.MAX_VALUE, mouseSpeedMax = new Double(0);
					BigDecimal mouseSpeedAve = new BigDecimal("0");
					Double mouseMovementDeviationMin = Double.MAX_VALUE, mouseMovementDeviationMax = new Double(0);
					BigDecimal mouseMovementDeviationAve = new BigDecimal("0");
					Double hesitationsMin = Double.MAX_VALUE, hesitationsMax = new Double(0);
					BigDecimal hesitationsAve = new BigDecimal("0");
					
					int movementLength = 0;
					int hesitationCounter = 0;
					int movementCounter = 0;
					List<MouseMovement> microMovementLog = new ArrayList<MouseMovement>();
					
					if (mouseFullLog != null && mouseFullLog.size() != 0)
					{
						for (Event event: mouseFullLog) {
							if (event instanceof MouseMovement) {
								if (microMovementLog.isEmpty() || event.getInstant().minus(microMovementLog.get(microMovementLog.size()-1).getInstant()) > Config.mouseMovementLimit) {
									if (microMovementLog.size() >= Config.mouseMovementMinLength) {
										double speed = (double) movementLength / (microMovementLog.get(microMovementLog.size()-1).getInstant().minus(microMovementLog.get(0).getInstant()));
										if (speed < mouseSpeedMin)
											mouseSpeedMin = speed;
										if (speed > mouseSpeedMax)
											mouseSpeedMax = speed;
										mouseSpeedAve = mouseSpeedAve.add(new BigDecimal(speed));
										
										double maxDistance = 0;
										for (MouseMovement pos: microMovementLog) {
											double dist = Math.abs(
												(microMovementLog.get(microMovementLog.size()-1).getX() - microMovementLog.get(0).getX()) * pos.getY() +
												(microMovementLog.get(0).getY() - microMovementLog.get(microMovementLog.size()-1).getY()) * pos.getX() +
												(microMovementLog.get(0).getX() - microMovementLog.get(microMovementLog.size()-1).getX()) * microMovementLog.get(0).getY() +
												(microMovementLog.get(microMovementLog.size()-1).getY() - microMovementLog.get(0).getY()) * microMovementLog.get(0).getX()
												) / Math.sqrt(Math.pow(microMovementLog.get(microMovementLog.size()-1).getX() - microMovementLog.get(0).getX(), 2) + Math.pow(microMovementLog.get(0).getY() - microMovementLog.get(microMovementLog.size()-1).getY(), 2));
											if (dist > maxDistance)
												maxDistance = dist;
										}
										if (maxDistance < mouseMovementDeviationMin)
											mouseMovementDeviationMin = maxDistance;
										if (maxDistance > mouseMovementDeviationMax)
											mouseMovementDeviationMax = maxDistance;
										mouseMovementDeviationAve = mouseMovementDeviationAve.add(new BigDecimal(maxDistance));
										
										if (hesitationCounter < hesitationsMin)
											hesitationsMin = (double) hesitationCounter;
										if (hesitationCounter > hesitationsMax)
											hesitationsMax = (double) hesitationCounter;
										hesitationsAve = hesitationsAve.add(new BigDecimal(hesitationCounter));
										
										movementCounter++;
									}
									movementLength = 0;
									hesitationCounter = 0;
									microMovementLog.clear();
								} else {
									movementLength += Math.sqrt(Math.pow(((MouseMovement) event).getX() - microMovementLog.get(microMovementLog.size()-1).getX(), 2) + Math.pow(((MouseMovement) event).getY() - microMovementLog.get(microMovementLog.size()-1).getY(), 2));
									
									if (!microMovementLog.isEmpty() && event.getInstant().minus(microMovementLog.get(microMovementLog.size()-1).getInstant()) > Config.mouseHesitationThreshold)
										hesitationCounter++;
								}
								microMovementLog.add((MouseMovement) event);
							} else {
								if (microMovementLog.size() >= Config.mouseMovementMinLength) {
									double speed = (double) movementLength / (microMovementLog.get(microMovementLog.size()-1).getInstant().minus(microMovementLog.get(0).getInstant()));
									if (speed < mouseSpeedMin)
										mouseSpeedMin = speed;
									if (speed > mouseSpeedMax)
										mouseSpeedMax = speed;
									mouseSpeedAve = mouseSpeedAve.add(new BigDecimal(speed));
									
									double maxDistance = 0;
									for (MouseMovement pos: microMovementLog) {
										double dist = Math.abs(
											(microMovementLog.get(microMovementLog.size()-1).getX() - microMovementLog.get(0).getX()) * pos.getY() +
											(microMovementLog.get(0).getY() - microMovementLog.get(microMovementLog.size()-1).getY()) * pos.getX() +
											(microMovementLog.get(0).getX() - microMovementLog.get(microMovementLog.size()-1).getX()) * microMovementLog.get(0).getY() +
											(microMovementLog.get(microMovementLog.size()-1).getY() - microMovementLog.get(0).getY()) * microMovementLog.get(0).getX()
											) / Math.sqrt(Math.pow(microMovementLog.get(microMovementLog.size()-1).getX() - microMovementLog.get(0).getX(), 2) + Math.pow(microMovementLog.get(0).getY() - microMovementLog.get(microMovementLog.size()-1).getY(), 2));
										if (dist > maxDistance)
											maxDistance = dist;
									}
									if (maxDistance < mouseMovementDeviationMin)
										mouseMovementDeviationMin = maxDistance;
									if (maxDistance > mouseMovementDeviationMax)
										mouseMovementDeviationMax = maxDistance;
									mouseMovementDeviationAve = mouseMovementDeviationAve.add(new BigDecimal(maxDistance));
									
									if (hesitationCounter < hesitationsMin)
										hesitationsMin = (double) hesitationCounter;
									if (hesitationCounter > hesitationsMax)
										hesitationsMax = (double) hesitationCounter;
									hesitationsAve = hesitationsAve.add(new BigDecimal(hesitationCounter));
									
									movementCounter++;
								}
								movementLength = 0;
								hesitationCounter = 0;
								microMovementLog.clear();
							}
						}
						
						if (microMovementLog.size() >= Config.mouseMovementMinLength) {
							double speed = (double) movementLength / (microMovementLog.get(microMovementLog.size()-1).getInstant().minus(microMovementLog.get(0).getInstant()));
							if (speed < mouseSpeedMin)
								mouseSpeedMin = speed;
							if (speed > mouseSpeedMax)
								mouseSpeedMax = speed;
							mouseSpeedAve = mouseSpeedAve.add(new BigDecimal(speed));
							
							double maxDistance = 0;
							for (MouseMovement pos: microMovementLog) {
								double dist = Math.abs(
									(microMovementLog.get(microMovementLog.size()-1).getX() - microMovementLog.get(0).getX()) * pos.getY() +
									(microMovementLog.get(0).getY() - microMovementLog.get(microMovementLog.size()-1).getY()) * pos.getX() +
									(microMovementLog.get(0).getX() - microMovementLog.get(microMovementLog.size()-1).getX()) * microMovementLog.get(0).getY() +
									(microMovementLog.get(microMovementLog.size()-1).getY() - microMovementLog.get(0).getY()) * microMovementLog.get(0).getX()
									) / Math.sqrt(Math.pow(microMovementLog.get(microMovementLog.size()-1).getX() - microMovementLog.get(0).getX(), 2) + Math.pow(microMovementLog.get(0).getY() - microMovementLog.get(microMovementLog.size()-1).getY(), 2));
								if (dist > maxDistance)
									maxDistance = dist;
							}
							if (maxDistance < mouseMovementDeviationMin)
								mouseMovementDeviationMin = maxDistance;
							if (maxDistance > mouseMovementDeviationMax)
								mouseMovementDeviationMax = maxDistance;
							mouseMovementDeviationAve = mouseMovementDeviationAve.add(new BigDecimal(maxDistance));
							
							if (hesitationCounter < hesitationsMin)
								hesitationsMin = (double) hesitationCounter;
							if (hesitationCounter > hesitationsMax)
								hesitationsMax = (double) hesitationCounter;
							hesitationsAve = hesitationsAve.add(new BigDecimal(hesitationCounter));
							
							movementCounter++;
						}
						if (movementCounter > 0) {
							mouseSpeedAve = mouseSpeedAve.divide(new BigDecimal(movementCounter), RoundingMode.HALF_EVEN);
							mouseMovementDeviationAve = mouseMovementDeviationAve.divide(new BigDecimal(movementCounter), RoundingMode.HALF_EVEN);
							hesitationsAve = hesitationsAve.divide(new BigDecimal(movementCounter), RoundingMode.HALF_EVEN);
							featureSet.setFeature(FeatureSet.Feature.MouseSpeedMin, mouseSpeedMin * 1000000);
							featureSet.setFeature(FeatureSet.Feature.MouseSpeedAve, mouseSpeedAve.doubleValue() * 1000000);
							featureSet.setFeature(FeatureSet.Feature.MouseSpeedMax, mouseSpeedMax * 1000000);
							featureSet.setFeature(FeatureSet.Feature.MouseMovementDeviationMin, mouseMovementDeviationMin);
							featureSet.setFeature(FeatureSet.Feature.MouseMovementDeviationAve, mouseMovementDeviationAve.doubleValue());
							featureSet.setFeature(FeatureSet.Feature.MouseMovementDeviationMax, mouseMovementDeviationMax);
							featureSet.setFeature(FeatureSet.Feature.MouseHesitationsPerMovementMin, hesitationsMin);
							featureSet.setFeature(FeatureSet.Feature.MouseHesitationsPerMovementAve, hesitationsAve.doubleValue());
							featureSet.setFeature(FeatureSet.Feature.MouseHesitationsPerMovementMax, hesitationsMax);
						}
					}
					return;
				}
			}
		}
	}
}
